@@ -1,41 +1,40 @@
-#!/usr/bin/env python3
-
 """Tools to generate a Snakemake-based BIDS app."""
 
-import os
-import pathlib
-import subprocess
 import argparse
 import logging
+import os
+import pathlib
+import re
+import subprocess
 import sys
-import yaml
-import bids
+
 import snakemake
+from colorama import Fore
 from snakemake.io import load_configfile
 
+from snakebids.exceptions import ConfigError, RunError
+from snakebids.utils.output import (
+    prepare_output,
+    retrofit_output,
+    write_config_file,
+    write_output_mode,
+)
+
 # We define Path here in addition to pathlib to put both variables in globals()
 # This way, users specifying a path type in their config.yaml can indicate
 # either Path or pathlib.Path
 Path = pathlib.Path
 
-bids.config.set_option("extension_initial_dot", True)
 logger = logging.Logger(__name__)
 
 
-class ConfigError(Exception):
-    """Exception raised for errors with the Snakebids config."""
-
-    def __init__(self, msg):
-        self.msg = msg
-        Exception.__init__()
-
-
+# pylint: disable=too-few-public-methods,
 class KeyValue(argparse.Action):
     """Class for accepting key=value pairs in argparse"""
 
     # Constructor calling
     def __call__(self, parser, namespace, values, option_string=None):
-        setattr(namespace, self.dest, dict())
+        setattr(namespace, self.dest, {})
 
         for value in values:
             # split it into key and value
@@ -44,6 +43,7 @@ class KeyValue(argparse.Action):
             getattr(namespace, self.dest)[key] = value
 
 
+# pylint: disable=too-few-public-methods
 class SnakemakeHelpAction(argparse.Action):
     """Class for printing snakemake usage in argparse"""
 
@@ -69,6 +69,7 @@ def run(command, env=None):
 
     merged_env = os.environ
     merged_env.update(env)
+    # pylint: disable=consider-using-with
     process = subprocess.Popen(
         command,
         stdout=subprocess.PIPE,
@@ -83,21 +84,13 @@ def run(command, env=None):
         if line == "" and process.poll() is not None:
             break
     if process.returncode != 0:
-        raise Exception("Non zero return code: %d" % process.returncode)
+        raise Exception(f"Non zero return code: {process.returncode}")
 
-def get_time_hash():
-    """ currently unused """
-    import hashlib
-    import time
-
-    hash = hashlib.sha1()
-    hash.update(str(time.time()).encode('utf-8'))
-    return hash.hexdigest()[:8]
 
 def resolve_path(path_candidate):
     """Helper function to resolve any paths or list
-    of paths it's passed. Otherwise, returns the argument 
-    unchanged. 
+    of paths it's passed. Otherwise, returns the argument
+    unchanged.
 
     Parameters
     ----------
@@ -115,9 +108,10 @@ def resolve_path(path_candidate):
 
     if isinstance(path_candidate, os.PathLike):
         return path_candidate.resolve()
-    
+
     return path_candidate
 
+
 SNAKEFILE_CHOICES = [
     "Snakefile",
     "snakefile",
@@ -127,16 +121,22 @@ SNAKEFILE_CHOICES = [
 
 
 CONFIGFILE_CHOICES = [
-            "config/snakebids.yml",
-            "config/snakebids.json",
-            "snakebids.yml",
-            "snakebids.json",
-            "config.yml",
-            "config.json",
-            "config/config.json",
-            "config/config.yml"]
+    "config/snakebids.yml",
+    "config/snakebids.yaml",
+    "config/snakebids.json",
+    "snakebids.yml",
+    "snakebids.yaml",
+    "snakebids.json",
+    "config.yml",
+    "config.yaml",
+    "config.json",
+    "config/config.json",
+    "config/config.yml",
+    "config/config.yaml",
+]
 
 
+# pylint: disable=too-few-public-methods, too-many-instance-attributes
 class SnakeBidsApp:
     """Snakebids app with config and arguments.
 
@@ -149,7 +149,7 @@ class SnakeBidsApp:
         If true, the Snakebids app will not attempt to parse input arguments,
         and will only handle the config file.
     out_configfile : str
-        Path to the updated configfile (YAML or JSON), relative to the output 
+        Path to the updated configfile (YAML or JSON), relative to the output
         working directory. This should be the same as the `configfile: ` used
         in your workflow. (default: 'config/snakebids.yml')
 
@@ -170,10 +170,21 @@ class SnakeBidsApp:
     configfile_path : str
         Relative path to config file (relative to snakemake_dir)
     updated_config : str
-        Absolute path to the updated config file to write
-
+        Absolute path to the updated config file to write.
+    workflow_mode : bool
+        If true, produces output in workflow mode, otherwise snakemake mode.
+    force : bool
+        If false, prevents conversion of output from workflow mode to bidsapp mode.
+    outputdir : Path
+        Path to outputdir specified by user
+    retrofit : bool
+        If true, run_snakemake will attempt to convert legacy output format into bidsapp
+        format.
     """
 
+    outputdir: Path
+    snakemake_dir: Path
+
     def __init__(self, snakemake_dir, skip_parse_args=False):
         # input argument is the dir where snakemake would be run
         # we use this to locate the config file, and snakefile adding them to
@@ -181,13 +192,15 @@ class SnakeBidsApp:
         # generated_config, so the workflow can use it to source files from
         # it (e.g. atlases etc..)
 
+        self.snakemake_dir = Path(snakemake_dir).resolve()
+
         # look for snakebids.yml in the snakemake_dir, quit if not found
-        self.configfile_path = None
+        self.configfile_path = ""
         for path in CONFIGFILE_CHOICES:
-            if Path(snakemake_dir, path).exists():
+            if (self.snakemake_dir / path).exists():
                 self.configfile_path = path
                 break
-        if self.configfile_path is None:
+        if not self.configfile_path:
             raise ConfigError(
                 f"Error: no config file found, tried {', '.join(CONFIGFILE_CHOICES)}."
             )
@@ -195,7 +208,7 @@ class SnakeBidsApp:
         # look for snakefile in the snakemake_dir, quit if not found
         self.snakefile = None
         for snakefile_path in SNAKEFILE_CHOICES:
-            if Path(snakemake_dir, snakefile_path).exists():
+            if (self.snakemake_dir / snakefile_path).exists():
                 self.snakefile = Path(snakemake_dir, snakefile_path)
                 break
         if self.snakefile is None:
@@ -203,8 +216,7 @@ class SnakeBidsApp:
                 f"Error: no Snakefile found, tried {', '.join(SNAKEFILE_CHOICES)}."
             )
 
-        self.config = load_configfile(Path(snakemake_dir,
-                                            self.configfile_path))
+        self.config = load_configfile(Path(snakemake_dir, self.configfile_path))
 
         if self.config.get("debug", False):
             logging.basicConfig(level=logging.DEBUG)
@@ -214,16 +226,14 @@ class SnakeBidsApp:
         self.config["snakemake_dir"] = snakemake_dir
         self.config["snakefile"] = self.snakefile
 
-        self.parser_include_snakemake = self.__create_parser(
-            include_snakemake=True
-        )
-        self.parser = self.__create_parser()
+        self.parser_include_snakemake = self._create_parser(include_snakemake=True)
+        self.parser = self._create_parser()
 
         if not skip_parse_args:
-            self.__parse_args()
+            self._parse_args()
 
-
-    def __create_parser(self, include_snakemake=False):
+    # pylint: disable=too-many-locals
+    def _create_parser(self, include_snakemake=False):
         """Create a parser with snakemake parser as parent solely for
         displaying help and checking conflicts, but then for actual parsing
         use snakebids parser to parse known args, then pass remaining to
@@ -245,8 +255,44 @@ class SnakeBidsApp:
                 description="Snakebids helps build BIDS Apps with Snakemake"
             )
 
+        parser.add_argument(
+            "--workflow-mode",
+            "--workflow_mode",
+            "-W",
+            action="store_true",
+            help=(
+                "Run Snakebids in Workflow mode. This will cause the entire Snakebids "
+                "app, except for the results folder, to be copied into your "
+                "output_dir. Snakemake will be run in this new child app, and results "
+                "will be put in output_dir/results."
+            ),
+        )
+
+        # We use -x as the alias because both -f and -F are taken by snakemake
+        parser.add_argument(
+            "--force-conversion",
+            "--force_conversion",
+            "-x",
+            action="store_true",
+            help=(
+                "Force snakebids to convert a workflow output to a bidsapp output. "
+                "conversion will delete all the files in the workflow snakebids app."
+            ),
+        )
+
+        parser.add_argument(
+            "--retrofit",
+            action="store_true",
+            help=(
+                "Convert a legacy Snakebids output (Snakebids 0.3.x or lower) into "
+                "bidsapp mode. This will delete any config files currently in the "
+                "output."
+            ),
+        )
+
         # add option for printing out snakemake usage
         parser.add_argument(
+            "--help-snakemake",
             "--help_snakemake",
             nargs=0,
             action=SnakemakeHelpAction,
@@ -257,9 +303,7 @@ class SnakeBidsApp:
         )
 
         # create parser group for app options
-        app_group = parser.add_argument_group(
-            "SNAKEBIDS", "Options for snakebids app"
-        )
+        app_group = parser.add_argument_group("SNAKEBIDS", "Options for snakebids app")
 
         # update the parser with config options
         for name, parse_args in self.config["parse_args"].items():
@@ -275,8 +319,17 @@ class SnakeBidsApp:
                         f"{parse_args['type']} is not available "
                         + f"as a type for {name}"
                     ) from err
-                    
-            app_group.add_argument(name, **parse_args)
+
+            if re.match(r"^--", name):
+                name_part = re.match(r"^--(.+)$", name).group(1)
+                names = (
+                    name,
+                    re.sub(r"\_", "-", name),
+                    f"--{re.sub(r'-', '_', name_part)}",
+                )
+            else:
+                names = (name,)
+            app_group.add_argument(*names, **parse_args)
 
         # general parser for
         # --filter_{input_type} {key1}={value1} {key2}={value2}...
@@ -287,7 +340,7 @@ class SnakeBidsApp:
         )
 
         for input_type in self.config["pybids_inputs"].keys():
-            argname = f"--filter_{input_type}"
+            argnames = (f"--filter_{input_type}", f"--filter-{input_type}")
             arglist_default = [
                 f"{key}={value}"
                 for (key, value) in self.config["pybids_inputs"][input_type][
@@ -297,7 +350,7 @@ class SnakeBidsApp:
             arglist_default_string = " ".join(arglist_default)
 
             filter_opts.add_argument(
-                argname,
+                *argnames,
                 nargs="+",
                 action=KeyValue,
                 help=f"(default: {arglist_default_string})",
@@ -312,15 +365,14 @@ class SnakeBidsApp:
         )
 
         for input_type in self.config["pybids_inputs"].keys():
-            argname = f"--wildcards_{input_type}"
+            argnames = (f"--wildcards-{input_type}", f"--wildcards_{input_type}")
             arglist_default = [
-                f"{wc}"
-                for wc in self.config["pybids_inputs"][input_type]["wildcards"]
+                f"{wc}" for wc in self.config["pybids_inputs"][input_type]["wildcards"]
             ]
             arglist_default_string = " ".join(arglist_default)
 
             wildcards_opts.add_argument(
-                argname,
+                *argnames,
                 nargs="+",
                 help=f"(default: {arglist_default_string})",
             )
@@ -336,12 +388,12 @@ class SnakeBidsApp:
 
         # create path override parser
         for input_type in self.config["pybids_inputs"].keys():
-            argname = f"--path_{input_type}"
-            override_opts.add_argument(argname, default=None)
+            argnames = (f"--path-{input_type}", f"--path_{input_type}")
+            override_opts.add_argument(*argnames, default=None)
 
         return parser
 
-    def __parse_args(self):
+    def _parse_args(self):
 
         # use snakebids parser to parse the known arguments
         # will pass the rest of args when running snakemake
@@ -350,9 +402,12 @@ class SnakeBidsApp:
         args = all_args[0]
         snakemake_args = all_args[1]
         # resolve all path items to get absolute paths
-        args.__dict__ = {
-            k: resolve_path(v) for k, v in args.__dict__.items()
-        }
+        args.__dict__ = {k: resolve_path(v) for k, v in args.__dict__.items()}
+
+        self.workflow_mode = args.workflow_mode
+        self.force = args.force_conversion
+        self.outputdir = Path(args.output_dir).resolve()
+        self.retrofit = args.retrofit
 
         # add snakebids arguments to config
         self.config.update(args.__dict__)
@@ -376,9 +431,7 @@ class SnakeBidsApp:
         for input_type in self.config["pybids_inputs"].keys():
             wildcards_list = self.config[f"wildcards_{input_type}"]
             if wildcards_list is not None:
-                self.config["pybids_inputs"][input_type][
-                    "wildcards"
-                ] += wildcards_list
+                self.config["pybids_inputs"][input_type]["wildcards"] += wildcards_list
             del self.config[f"wildcards_{input_type}"]
 
         # add custom input paths to
@@ -386,80 +439,90 @@ class SnakeBidsApp:
         for input_type in self.config["pybids_inputs"].keys():
             custom_path = self.config[f"path_{input_type}"]
             if custom_path is not None:
-                self.config["pybids_inputs"][input_type][
-                    "custom_path"
-                ] = Path(custom_path).resolve()
+                self.config["pybids_inputs"][input_type]["custom_path"] = Path(
+                    custom_path
+                ).resolve()
             del self.config[f"path_{input_type}"]
 
         # replace paths with realpaths
         self.config["bids_dir"] = Path(self.config["bids_dir"]).resolve()
         self.config["output_dir"] = Path(self.config["output_dir"]).resolve()
 
-    
-    def write_updated_config(self):
-        """Create an updated snakebids config file in the output dir."""
-        self.updated_config = Path(self.config["output_dir"],
-                                    self.configfile_path)
-
-
-        # create the output folder if needed
-        self.updated_config.parent.mkdir(parents = True, exist_ok=True)
-
-
-        time_hash = get_time_hash() # TODO: copy to a time-hashed file too
-                                    # for provenance? 
-                                    #  unused as of now.. 
-
-
-        with open(self.updated_config, "w") as f:
-            # write either as JSON or YAML
-            if self.updated_config.suffix == '.json':
-                import json
-                json.dump(self.config, f, indent=4)
-            else: #if not json, then should be yaml or yml
-                from collections import OrderedDict
-                
-                #this is needed to make the output yaml clean
-                yaml.add_representer(OrderedDict,
-                                        lambda dumper,data: 
-                                            dumper.represent_mapping(
-                                                'tag:yaml.org,2002:map',
-                                                data.items()))
-
-                # Represent any PathLikes as str.
-                path2str = lambda dumper, data: dumper.represent_scalar('tag:yaml.org,2002:str',str(data))
-                yaml.add_representer(pathlib.PosixPath, path2str)
-                yaml.add_representer(pathlib.WindowsPath, path2str)
-
-                yaml.dump(dict(self.config),
-                            f, 
-                            default_flow_style=False,
-                            sort_keys=False)
-
     def run_snakemake(self):
-        """Run snake make with that config.
+        """Run snakemake with that config.
 
         Workflow snakefile will read snakebids config, create inputs_config,
         and read that in.
         """
 
-        # write updated config
-        self.write_updated_config()
+        if self.snakemake_dir == self.outputdir:
+            write_output_mode(self.outputdir / ".snakebids", "workflow")
+            mode = "workflow"
+            force_config_overwrite = True
+            if not self.workflow_mode:
+                print(
+                    f"{Fore.YELLOW}You specified your output to be in the snakebids "
+                    "directory, so we're switching automatically to workflow mode!\n"
+                    f"{Fore.RESET}You'll find your results in "
+                    f"`{(self.snakemake_dir/'results').resolve()}`"
+                )
+        else:
+            mode = "workflow" if self.workflow_mode else "bidsapp"
+            force_config_overwrite = False
+            if self.retrofit:
+                for path in CONFIGFILE_CHOICES:
+                    if (self.outputdir / path).exists():
+                        self.configfile_path = path
+                if not retrofit_output(
+                    self.outputdir,
+                    # Find all config files in the outputdir
+                    (
+                        self.outputdir / p
+                        for p in CONFIGFILE_CHOICES
+                        if (self.outputdir / p).exists()
+                    ),
+                ):
+                    print(f"{Fore.YELLOW}Exiting. No conversion performed.{Fore.RESET}")
+                    sys.exit(1)
+
+        try:
+            root = prepare_output(self.snakemake_dir, self.outputdir, mode, self.force)
+        except RunError as err:
+            print(err.msg)
+            sys.exit(1)
+
+        if mode == "workflow":
+            self.config["output_dir"] = str(root)
+            self.config["root"] = "results"
+            new_config_file = self.outputdir / self.configfile_path
+        else:
+            force_config_overwrite = True
+            self.config["root"] = ""
+            new_config_file = self.outputdir / self.configfile_path
+
+        cwd = self.outputdir
+
+        write_config_file(new_config_file, self.config, force_config_overwrite)
 
         # running the chosen participant level
-
         analysis_level = self.config["analysis_level"]
-        # runs snakemake, using the workflow config and inputs config to
-        # override
 
-        # run snakemake command-line (passing any leftover args from argparse)
-        snakemake_cmd_list = [
-            "snakemake",
-            f"--snakefile {self.snakefile}",
-            f"--directory {self.config['output_dir']}",
-            *self.config["snakemake_args"],
-            *self.config["targets_by_analysis_level"][analysis_level],
+        # run snakemake (passing any leftover args from argparse)
+        # Filter any blank strings before submitting
+        snakemake_argv = [
+            *filter(
+                None,
+                [
+                    "--snakefile",
+                    str(self.snakefile),
+                    "--directory",
+                    str(cwd),
+                    "--configfile",
+                    str(new_config_file.resolve()),
+                    *self.config["snakemake_args"],
+                    *self.config["targets_by_analysis_level"][analysis_level],
+                ],
+            )
         ]
 
-        snakemake_cmd = " ".join(snakemake_cmd_list)
-        run(snakemake_cmd)
+        snakemake.main(snakemake_argv)
